# -*- coding: utf-8 -*-

'''
Support SQLite storage format
'''

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.


import logging

from chirptext import DataObject
from chirptext import ttl
from puchikarui import Schema, with_ctx
from .data import INIT_TTL_SQLITE


# ----------------------------------------------------------------------
# Configuration
# ----------------------------------------------------------------------

def getLogger():
    return logging.getLogger(__name__)


# ----------------------------------------------------------------------
# Models
# ----------------------------------------------------------------------

class Meta(DataObject):
    def __init__(self, key='', value='', **kwargs):
        super().__init__(**kwargs)
        self.key = key
        self.value = value

    def __repr__(self):
        data = ("{}={}".format(k, v) for k, v in self.to_dict().items())
        return "Meta({})".format(", ".join(data))

    def __str__(self):
        return str(self.to_dict())


class DocMeta(DataObject):
    def __repr__(self):
        data = ("{}={}".format(k, v) for k, v in self.to_dict().items())
        return "DocMeta({})".format(", ".join(data))

    def __str__(self):
        return str(self.to_dict())


class CorpusMeta(DataObject):
    def __repr__(self):
        data = ("{}={}".format(k, v) for k, v in self.to_dict().items())
        return "CorpusMeta({})".format(", ".join(data))

    def __str__(self):
        return str(self.to_dict())


class Corpus(DataObject):
    pass


class CWLink(DataObject):
    pass


class TTLSQLite(Schema):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_file(INIT_TTL_SQLITE)
        # add tables
        self.add_table('meta', ['key', 'value'], proto=Meta).set_id('key')
        self.add_table('meta_doc', ['name', 'key', 'value'], proto=DocMeta)
        self.add_table('meta_cor', ['name', 'key', 'value'], proto=CorpusMeta)
        self.add_table('corpus', ['ID', 'name', 'title'], proto=Corpus).set_id('ID')
        self.add_table('document', ['ID', 'name', 'title', 'lang', 'corpusID'],
                       proto=ttl.Document, alias='doc').set_id('ID')
        self.add_table('sentence', ['ID', 'ident', 'text', 'docID', 'flag', 'comment'],
                       proto=ttl.Sentence, alias='sent').set_id('ID')
        self.add_table('token', ['ID', 'sid', 'widx', 'text', 'lemma', 'pos', 'cfrom', 'cto', 'comment'],
                       proto=ttl.Token).set_id('ID')
        self.add_table('concept', ['ID', 'sid', 'cidx', 'clemma', 'tag', 'flag', 'comment'],
                       proto=ttl.Concept).set_id('ID')
        self.add_table('tag', ['ID', 'sid', 'wid', 'cfrom', 'cto', 'label', 'source', 'tagtype'],
                       proto=ttl.Tag).set_id('ID')
        self.add_table('cwl', ['sid', 'cid', 'wid'], proto=CWLink)

    @with_ctx
    def new_corpus(self, name, title='', ctx=None):
        corpus = Corpus(name=name, title=title)
        newid = ctx.corpus.save(corpus)
        corpus.ID = newid
        return corpus

    @with_ctx
    def new_doc(self, name, corpusID, title='', lang='', ctx=None, **kwargs):
        doc = ttl.Document(name=name, corpusID=corpusID, title=title, lang=lang, **kwargs)
        newid = ctx.doc.save(doc)
        doc.ID = newid
        return doc

    @with_ctx
    def ensure_corpus(self, name, ctx=None, **kwargs):
        corpus = ctx.corpus.select_single('name=?', (name,))
        if corpus is None:
            corpus = self.new_corpus(name, ctx=ctx, **kwargs)
        return corpus

    @with_ctx
    def ensure_doc(self, name, corpus, ctx=None, **kwargs):
        doc = ctx.doc.select_single('name = ?', (name,))
        if doc is None:
            doc = self.new_doc(name=name, corpusID=corpus.ID, ctx=ctx, **kwargs)
        return doc

    def simplify_tag(self, a_tag):
        if a_tag.cfrom == -1:
            a_tag.cfrom = None
        if a_tag.cto == -1:
            a_tag.cto = None
        if not a_tag.source:
            a_tag.source = None
        return a_tag

    @with_ctx
    def save_sent(self, sent_obj, ctx=None):
        # insert sentence
        # save sent obj first
        sent_obj.ID = ctx.sent.save(sent_obj)
        # save sentence's tags
        for tag in sent_obj.tags:
            tag.sid = sent_obj.ID
            tag.wid = None  # ensure that wid is not saved
            self.simplify_tag(tag)
            tag.ID = ctx.tag.save(tag)
        # save tokens
        for idx, token in enumerate(sent_obj):
            token.sid = sent_obj.ID
            token.widx = idx
            token.ID = ctx.token.save(token)
            # save token's tags
            for tag in token:
                tag.sid = sent_obj.ID
                tag.wid = token.ID
                self.simplify_tag(tag)
                tag.ID = ctx.tag.save(tag)
        # save concepts
        for idx, concept in enumerate(sent_obj.concepts):
            concept.sid = sent_obj.ID
            concept.ID = ctx.concept.save(concept)
            # save cwl
            for token in concept.tokens:
                cwl = CWLink(sid=sent_obj.ID, cid=concept.ID, wid=token.ID)
                ctx.cwl.save(cwl)
        return sent_obj

    @with_ctx
    def get_sent(self, sentID, ctx=None):
        sent = ctx.sent.by_id(sentID)
        # select tokens
        tokens = ctx.token.select('sid = ?', (sent.ID,))
        tokenmap = {t.ID: t for t in tokens}
        for tk in tokens:
            sent.tokens.append(tk)
        # select all tags
        tags = ctx.tag.select('sid = ?', (sent.ID,))
        for tag in tags:
            if tag.wid is None:
                sent.tags.append(tag)
            elif tag.wid in tokenmap:
                tokenmap[tag.wid].tags.append(tag)
            else:
                getLogger().warning("Orphan tag in sentence #{}: {}".format(sent.ID, tag))
        # select concepts
        concepts = ctx.concept.select('sid = ?', (sent.ID,))
        conceptmap = {c.ID: c for c in concepts}
        for c in concepts:
            sent.add_concept(c)
        # select cwl
        cwlinks = ctx.cwl.select('sid = ?', (sent.ID,))
        for cwl in cwlinks:
            conceptmap[cwl.cid].add_token(tokenmap[cwl.wid])
        return sent

    @with_ctx
    def lexicon(self, limit=None, ctx=None):
        query = 'SELECT text, COUNT(*) FROM token GROUP BY text ORDER BY COUNT(*) DESC'
        params = []
        if limit:
            query += ' LIMIT ?'
            params.append(limit)
        return ctx.execute(query, params)

    # ---- Meta related functions
    @with_ctx
    def get_meta(self, ctx=None):
        return ctx.meta.select()

    @with_ctx
    def get_meta_by_key(self, key, ctx=None):
        return ctx.meta.by_id(key)

    @with_ctx
    def set_meta(self, key, value, ctx=None):
        query = '''INSERT OR REPLACE INTO meta VALUES (?, ?)'''
        params = (key, value)
        return ctx.execute(query, params)

    @with_ctx
    def get_doc_meta(self, name, ctx=None):
        return ctx.meta_doc.select('name = ?', (name,))

    @with_ctx
    def get_doc_meta_by_key(self, name, key, ctx=None):
        return ctx.meta_doc.select_single('name = ? and key = ?', (name, key))

    @with_ctx
    def set_doc_meta(self, name, key, value, ctx=None):
        query = '''INSERT OR REPLACE INTO meta_doc VALUES (?, ?, ?)'''
        params = (name, key, value)
        return ctx.execute(query, params)

    @with_ctx
    def get_cor_meta(self, name, ctx=None):
        return ctx.meta_cor.select('name = ?', (name,))

    @with_ctx
    def get_cor_meta_by_key(self, name, key, ctx=None):
        return ctx.meta_cor.select_single('name = ? and key = ?', (name, key))

    @with_ctx
    def set_cor_meta(self, name, key, value, ctx=None):
        query = '''INSERT OR REPLACE INTO meta_cor VALUES (?, ?, ?)'''
        params = (name, key, value)
        return ctx.execute(query, params)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Test ORG-mode
"""

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import os
import io
import unittest
import logging

from speach import orgmode


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

TEST_DIR = os.path.dirname(os.path.realpath(__file__))
TEST_ORG = '''#+TITLE: まだらの紐
- author :: アーサー・コナン・ドイル
- genre :: 短編小説

同様に、悲しみそのものを、それが悲しみであるという理由で愛する者や、それゆえ得ようとする者は、どこにもいない。
同様に、悲しみそのものを、それが悲しみであるという理由で愛する者や、それゆえ得ようとする者は、どこにもいない。
同様に、悲しみそのものを、それが悲しみであるという理由で愛する者や、それゆえ得ようとする者は、どこにもいない。
'''


def getLogger():
    return logging.getLogger(__name__)


# -------------------------------------------------------------------------------
# Tests
# -------------------------------------------------------------------------------


class TestOrgMode(unittest.TestCase):

    def test_title(self):
        title = orgmode._match_title('#+TITLE: まだらの紐\n')
        self.assertEqual(title, 'まだらの紐')

    def test_meta(self):
        k, v = orgmode._match_meta('- genre :: fiction')
        self.assertEqual((k, v), ('genre', 'fiction'))

    def test_parse(self):
        instream = io.StringIO(TEST_ORG)
        t, m, l = orgmode._parse_stream(instream)
        self.assertEqual(t, 'まだらの紐')
        self.assertEqual(m, [('author', 'アーサー・コナン・ドイル'), ('genre', '短編小説')])
        self.assertEqual(len(l), 3)


# -------------------------------------------------------------------------------
# MAIN
# -------------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

'''
Web Video Text Tracks format (WebVTT) support

More information:
    WebVTT: The Web Video Text Tracks Format
        https://www.w3.org/2013/07/webvtt.html
'''

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.


import re
import math

# ----------------------------------------------------------------------
# Configuration
# ----------------------------------------------------------------------

WEBVTT = re.compile(r'(?P<hour>\d{2,}):(?P<min>\d{2}):(?P<sec>\d{2}.\d{3})')


# ----------------------------------------------------------------------
# Models
# ----------------------------------------------------------------------

def sec2ts(seconds: float) -> str:
    ''' Convert duration in seconds in seconds to VTT format (e.g. 01:53:47.262)

    >>> from speach import vtt
    >>> vtt.sec2ts(25.403)
    '00:00:25.403'
    >>> vtt.sec2ts(14532.768)
    '04:02:12.768'

    :param seconds: Timestamp in seconds
    :type ts: float
    :returns: Web VTT formatted timestamp
    :rtype: str
    :raises: :exc:`ValueError`: when seconds is not a number or cannot be casted into a number
    '''
    try:
        seconds = float(seconds)
        if seconds < 0:
            raise ValueError("Timestamp cannot be smaller than 0")
    except Exception as e:
        raise ValueError("Invalid timestamp ({})".format(seconds)) from e
    min_base = math.floor(seconds / 60)
    ts_hour = math.floor(min_base / 60)
    ts_min = math.floor(min_base % 60)
    ts_sec = math.floor(seconds) % 60
    ts_msec = (seconds - int(seconds)) * 1000
    ts = "{:0>2.0f}:{:0>2.0f}:{:0>2.0f}.{:0>3.0f}".format(ts_hour, ts_min, ts_sec, ts_msec)
    return ts


def ts2sec(ts: str) -> float:
    ''' Convert VTT timestamp to duration (float, in seconds)

    >>> from speach import vtt
    >>> vtt.ts2sec("01:00:41.231")
    3641.231
    >>> print(vtt.ts2sec("00:00:00.000"))
    0.0

    :param ts: Timestamp (hh:mm:ss.nnn)
    :type ts: str
    :returns: timestamp in seconds
    :rtype: float
    :raises: :exc:`ValueError`: if timestamp is not formatted correctly
    '''
    if ts is None:
        raise ValueError("Timestamp cannot be empty")
    m = WEBVTT.match(ts)
    if not m:
        raise ValueError("Invalid VTT timestamp format ({})".format(ts))
    pd = m.groupdict()  # parts' dictionary
    secs = int(pd['hour']) * 3600
    secs += int(pd['min']) * 60
    secs += float(pd['sec'])
    return secs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Test TTL Interlinear Gloss (TIG)
'''

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import os
import io
import unittest
import logging
from collections import OrderedDict

from chirptext import chio
from chirptext import deko

from speach import ttl
from speach import ttlig
from speach.ttlig import IGStreamReader, TTLTokensParser


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

TEST_DIR = os.path.dirname(os.path.realpath(__file__))
JP_IMPLICIT = os.path.join(TEST_DIR, 'data', 'testig_jp_implicit.txt')
JP_EXPLICIT = os.path.join(TEST_DIR, 'data', 'testig_jp_explicit.txt')
JP_MANUAL = os.path.join(TEST_DIR, 'data', 'testig_jp_manual.txt')
VN_EXPLICIT = os.path.join(TEST_DIR, 'data', 'testig_vi_explicit.txt')
TRANSCRIPT_FILE = os.path.join(TEST_DIR, 'data', 'test_transcript.tab')
TRANSCRIPT_EXPECTED_FILE = os.path.join(TEST_DIR, 'data', 'test_transcript.human.tab')


def getLogger():
    return logging.getLogger(__name__)


# -------------------------------------------------------------------------------
# Tests
# -------------------------------------------------------------------------------

class TestTokenizer(unittest.TestCase):

    def test_tokenizer(self):
        print("test multiple spaces for tokenizing")
        sent = ttl.Sentence('It works.')
        token_string = 'It       works   .    '
        gloss_string = 'SUBJ     work    PUNC '
        tokens = ttlig.tokenize(token_string)
        glosses = ttlig.tokenize(gloss_string)
        sent.tokens = tokens
        for tk, gl in zip(sent.tokens, glosses):
            tk.new_tag(gl, tagtype='gloss')
        # verify imported information
        actual = [(t.text, t.get_tag('gloss').label) for t in sent]
        expected = [('It', 'SUBJ'), ('works', 'work'), ('.', 'PUNC')]
        self.assertEqual(expected, actual)

    def test_tokenizing_special_chars(self):
        ''' Only 2 characters - escapechar and delimiter '''
        tokens = ttlig.tokenize('some\\ thing is a word .')
        expected = ['some thing', 'is', 'a', 'word', '.']
        self.assertEqual(tokens, expected)
        # last char is special
        tokens = ttlig.tokenize('some\\ thing is a word \\.')
        expected = ['some thing', 'is', 'a', 'word', '.']
        self.assertEqual(tokens, expected)
        # last char is delimiter
        self.assertRaises(Exception, lambda: ttlig.tokenize('this is wrong\\'))


class TestFurigana(unittest.TestCase):

    def test_parse_furigana(self):
        rubytext = ttlig.parse_furigana('')
        self.assertEqual(str(rubytext), '')
        self.assertEqual(rubytext.to_html(), '')

        self.assertRaises(ValueError, lambda: ttlig.parse_furigana(None))

        rubytext = ttlig.parse_furigana('{食/た}べる')
        self.assertEqual(str(rubytext), '食べる')
        self.assertEqual(rubytext.to_html(), '<ruby><rb>食</rb><rt>た</rt></ruby>べる')

        rubytext = ttlig.parse_furigana('{面/おも}{白/しろ}い')
        self.assertEqual(str(rubytext), '面白い')
        self.assertEqual(rubytext.to_html(), '<ruby><rb>面</rb><rt>おも</rt></ruby><ruby><rb>白</rb><rt>しろ</rt></ruby>い')

        rubytext = ttlig.parse_furigana('{漢字/かんじ}')
        self.assertEqual(str(rubytext), '漢字')
        self.assertEqual(rubytext.to_html(), '<ruby><rb>漢字</rb><rt>かんじ</rt></ruby>')

        rubytext = ttlig.parse_furigana('お{天/てん}{気/き}')
        self.assertEqual(str(rubytext), 'お天気')
        self.assertEqual(rubytext.to_html(), 'お<ruby><rb>天</rb><rt>てん</rt></ruby><ruby><rb>気</rb><rt>き</rt></ruby>')
        # weird cases
        rubytext = ttlig.parse_furigana('{{漢字/かんじ}}')
        self.assertEqual(str(rubytext), '{漢字}')
        self.assertEqual(rubytext.to_html(), '{<ruby><rb>漢字</rb><rt>かんじ</rt></ruby>}')
        # %
        rubytext = ttlig.parse_furigana('{％/パーセント}')
        self.assertEqual(str(rubytext), '％')
        self.assertEqual(rubytext.to_html(), '<ruby><rb>％</rb><rt>パーセント</rt></ruby>')

        rubytext = ttlig.parse_furigana('お{天/てん{気/き}')
        self.assertEqual(str(rubytext), 'お{天/てん気')  # first one won't be matched
        self.assertEqual(rubytext.to_html(), 'お{天/てん<ruby><rb>気</rb><rt>き</rt></ruby>')

    def test_TTL_tokenizer(self):
        parser = TTLTokensParser()
        tokens = parser.parse_ruby('{猫/ねこ} が {好/す}き です 。')
        token_text = [t.text() for t in tokens]
        self.assertEqual(token_text, ['猫', 'が', '好き', 'です', '。'])
        actual = parser.delimiter.join(r.to_html() for r in tokens)
        expected = '<ruby><rb>猫</rb><rt>ねこ</rt></ruby> が <ruby><rb>好</rb><rt>す</rt></ruby>き です 。'
        self.assertEqual(expected, actual)
        # test parse_ruby
        actual = ttlig.make_ruby_html('ケーキ を {食/た}べた 。')
        expected = 'ケーキ を <ruby><rb>食</rb><rt>た</rt></ruby>べた 。'
        self.assertEqual(expected, actual)


class TestTTLIG(unittest.TestCase):

    def test_iter_stream(self):
        raw = io.StringIO('''# TTLIG
# This is a comment
I drink green tea.
I drink green_tea.

I have two cats.
I have two cat-s.
''')
        groups = [x for x in IGStreamReader._iter_stream(raw)]
        expected = [['I drink green tea.', 'I drink green_tea.'], ['I have two cats.', 'I have two cat-s.']]
        self.assertEqual(groups, expected)

        # nothing
        raw = io.StringIO('''# TTLIG''')
        groups = [x for x in IGStreamReader._iter_stream(raw)]
        expected = []
        self.assertEqual(groups, expected)
        raw = io.StringIO('')
        groups = [x for x in IGStreamReader._iter_stream(raw)]
        expected = []
        self.assertEqual(groups, expected)
        raw = io.StringIO('a\n\nb')
        groups = [x for x in IGStreamReader._iter_stream(raw)]
        expected = [['a'], ['b']]
        self.assertEqual(groups, expected)
        raw = io.StringIO('a\n#comment\nb')
        groups = [x for x in IGStreamReader._iter_stream(raw)]
        expected = [['a', 'b']]
        self.assertEqual(groups, expected)

    def test_read_header(self):
        inpath = VN_EXPLICIT
        with chio.open(inpath) as infile:
            meta = ttlig.IGStreamReader._read_header(infile)
        expected = OrderedDict([('Language', 'Vietnamese'), ('Language code', 'vie'), ('Lines', 'orth translit gloss translat'), ('Author', 'Le Tuan Anh'), ('Date', 'May 25 2018')])
        self.assertEqual(meta, expected)

    def test_read_file(self):
        inpath = JP_MANUAL
        s1, s2 = ttlig.read(inpath)
        s1_json = {'text': '猫が好きです。', 'transliteration': 'neko ga suki desu .', 'transcription': '', 'morphtrans': '', 'morphgloss': 'cat SUBM likeable COP .', 'wordgloss': '', 'translation': 'I like cats.', 'ident': '01a_01', 'tokens': '{猫/ねこ} が {好/す}き です 。'}
        s2_json = {'text': '雨が降る。', 'transliteration': '', 'transcription': '', 'morphtrans': '', 'morphgloss': '', 'wordgloss': '', 'translation': 'It rains.', 'ident': '01a_02', 'tokens': '{雨/あめ} が {降/ふ}る 。'}
        self.assertEqual(s1.to_dict(), s1_json)
        self.assertEqual(s2.to_dict(), s2_json)
        # test furigana
        s1_furi = '<ruby><rb>猫</rb><rt>ねこ</rt></ruby> が <ruby><rb>好</rb><rt>す</rt></ruby>き です 。'
        s2_furi = '<ruby><rb>雨</rb><rt>あめ</rt></ruby> が <ruby><rb>降</rb><rt>ふ</rt></ruby>る 。'
        self.assertEqual(ttlig.make_ruby_html(s1.tokens), s1_furi)
        self.assertEqual(ttlig.make_ruby_html(s2.tokens), s2_furi)

    def test_ttlig_auto(self):
        inpath = JP_IMPLICIT
        sents = ttlig.read(inpath)
        s = sents[0]
        self.assertEqual(s.text, '猫が好きです。')
        self.assertEqual(s.tokens, '{猫/ねこ} が {好/す}き です 。')
        self.assertEqual(s.morphgloss, 'cat SUBM likeable COP .')
        self.assertEqual(s.translation, 'I like cats.')

    def test_ttlig_manual(self):
        inpath = JP_MANUAL
        s1, s2 = ttlig.read(inpath)
        s1_dict = {'text': '猫が好きです。', 'transliteration': 'neko ga suki desu .', 'transcription': '', 'morphtrans': '', 'morphgloss': 'cat SUBM likeable COP .', 'wordgloss': '', 'translation': 'I like cats.', 'ident': '01a_01', 'tokens': '{猫/ねこ} が {好/す}き です 。'}
        s2_dict = {'text': '雨が降る。', 'transliteration': '', 'transcription': '', 'morphtrans': '', 'morphgloss': '', 'wordgloss': '', 'translation': 'It rains.', 'ident': '01a_02', 'tokens': '{雨/あめ} が {降/ふ}る 。'}
        self.assertEqual(s1.to_dict(), s1_dict)
        self.assertEqual(s2.to_dict(), s2_dict)

    def test_read_empty_file(self):
        instream = io.StringIO('# TTLIG')
        sents = ttlig.read_stream(instream)
        self.assertEqual(sents, [])

    def test_read_invalid_ttlig(self):
        instream = io.StringIO('')
        self.assertRaises(Exception, lambda: ttlig.read_stream(instream))
        invalid_file = os.path.join(TEST_DIR, 'data', 'testig_invalid.txt')
        self.assertRaises(Exception, lambda: ttlig.read(invalid_file))

    def test_make_furi_token(self):
        s = deko.parse('友達')
        # f = ttlig.mctoken_to_furi(s[0])
        f = ttlig.RubyToken.from_furi(s[0].surface, s[0].reading_hira())
        self.assertEqual(f.to_code(), '{友達/ともだち}')
        # half-width char
        s = deko.parse('0')
        f = ttlig.RubyToken.from_furi(s[0].surface, s[0].reading_hira())
        self.assertEqual(f.to_code(), '0')

    def test_parsing(self):
        igrow = ttlig.text_to_igrow('友達と巡り会った。')
        self.assertEqual(igrow.text, '友達と巡り会った。')
        self.assertEqual(igrow.tokens, '{友達/ともだち} と {巡/めぐ}り{会/あ}っ た 。')
        expected = '<ruby><rb>友達</rb><rt>ともだち</rt></ruby> と <ruby><rb>巡</rb><rt>めぐ</rt></ruby>り<ruby><rb>会</rb><rt>あ</rt></ruby>っ た 。'
        actual = ttlig.make_ruby_html(igrow.tokens)
        self.assertEqual(expected, actual)
        # test
        igrow = ttlig.text_to_igrow('言い尽くす')
        self.assertEqual(igrow.text, '言い尽くす')
        self.assertEqual(igrow.tokens, '{言/い}い{尽/つ}くす')
        # test more
        igrow = ttlig.text_to_igrow('言いなさい')
        self.assertEqual(igrow.text, '言いなさい')
        self.assertEqual(igrow.tokens, '{言/い}い なさい')
        # half-width char
        igrow = ttlig.text_to_igrow('0時だ。')
        self.assertEqual(igrow.text, '0時だ。')
        self.assertEqual(igrow.tokens, '0 {時/じ} だ 。')

    def test_parsing_aligned_text(self):
        print("Testing TTLIG with multiple spaces")


# -------------------------------------------------------------------------------
# MAIN
# -------------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

'''
org-mode support
'''

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.


import re
import os
import logging

from chirptext import chio


# ----------------------------------------------------------------------
# Configuration
# ----------------------------------------------------------------------

def getLogger():
    return logging.getLogger(__name__)


RE_TITLE = re.compile('#\+TITLE: (?P<title>.+)$')
RE_META = re.compile('- (?P<tag>.+) :: (?P<value>.+)$')


# ----------------------------------------------------------------------
# Models
# ----------------------------------------------------------------------

def _match_title(line):
    m = RE_TITLE.match(line)
    if m:
        return m.group('title')
    else:
        return None


def _match_meta(line):
    m = RE_META.match(line)
    if m:
        return (m.group('tag'), m.group('value'))
    else:
        return None


def _parse_stream(input_stream):
    title = None
    meta = []
    reading_header = True
    lines = []
    for idx, line in enumerate(input_stream):
        if idx == 0:
            m = _match_title(line)
            if m:
                title = m
                continue
        # not title
        if reading_header:
            m = _match_meta(line)
            if m:
                meta.append(m)
                continue
            else:
                # not a meta line
                reading_header = False
                if not line.strip():
                    # ignore the first empty line after meta lines
                    continue
        # add to lines
        lines.append(line)
    return (title, meta, lines)


def read(filepath, **kwargs):
    with chio.open(filepath, mode='r') as infile:
        title, meta, lines = _parse_stream(infile)
        meta.append(('Filename', os.path.basename(filepath)))
        for k, v in kwargs.items():
            meta.append((k, v))
    return (title, meta, lines)


def org_to_ttlig(title, meta, lines, line_processor=None):
    iglines = ['# TTLIG']
    # add title
    if title:
        iglines.append("Title: {}".format(title))
    # add meta
    for k, v in meta:
        iglines.append("{}: {}".format(k, v))
    # add an empty between meta and content
    if meta:
        iglines.append('')

    # add lines
    for line in lines:
        if line_processor:
            line_processor(line, iglines)
        else:
            iglines.append(line)
    return iglines



# -*- coding: utf-8 -*-

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

__author__ = "Le Tuan Anh"
__email__ = "tuananh.ke@gmail.com"
__copyright__ = "Copyright (c) 2018, Le Tuan Anh <tuananh.ke@gmail.com>"
__credits__ = []
__license__ = "MIT License"
__description__ = "a Python library for Managing, annotating, and converting natural language corpuses using popular formats (CoNLL, ELAN, Praat, CSV, JSON, SQLite, VTT, Audacity, TTL, TIG, ISF)"
__url__ = "https://github.com/neocl/speach/"
__maintainer__ = "Le Tuan Anh"
__version_major__ = "0.1"  # follow PEP-0440
__version__ = "{}a1".format(__version_major__)
__version_long__ = "{} - Alpha 1".format(__version_major__)
__status__ = "Prototype"



# -*- coding: utf-8 -*-

'''
ELAN module - manipulating ELAN transcript files (\*.eaf, \*.pfsx)
'''

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.


import logging
from collections import OrderedDict
from collections import defaultdict as dd
from typing import List, Tuple
import xml.etree.ElementTree as etree

from chirptext import DataObject
from chirptext import chio

from .vtt import sec2ts, ts2sec


# ----------------------------------------------------------------------
# Configuration
# ----------------------------------------------------------------------

def getLogger():
    return logging.getLogger(__name__)


# ----------------------------------------------------------------------
# Models
# ----------------------------------------------------------------------

CSVRow = List[str]
CSVTable = List[CSVRow]


class TimeSlot():
    def __init__(self, ID, value=None, xml_node=None, *args, **kwargs):
        """ An ELAN timestamp (with ID)
        """
        self.ID = ID
        self.value = value
        self.__xml_node = xml_node

    @property
    def ts(self):
        return sec2ts(self.sec) if self.value is not None else None

    @property
    def sec(self):
        return self.value / 1000 if self.value is not None else None

    def __lt__(self, other):
        if other is None or (isinstance(other, TimeSlot) and other.value is None):
            return False
        return self.value < other.value if isinstance(other, TimeSlot) else self.value < other

    def __eq__(self, other):
        if other is None or (isinstance(other, TimeSlot) and other.value is None):
            return False
        return self.value == other.value if isinstance(other, TimeSlot) else self.value == other

    def __gt__(self, other):
        if other is None or (isinstance(other, TimeSlot) and other.value is None):
            return True
        return self.value > other.value if isinstance(other, TimeSlot) else self.value > other

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return self > other or self == other

    def __add__(self, other):
        return self.value + other.value if isinstance(other, TimeSlot) else self.value + other

    def __sub__(self, other):
        return self.value - other.value if isinstance(other, TimeSlot) else self.value - other

    def __hash__(self):
        return id(self)

    def __str__(self):
        val = self.ts
        return val if val else self.ID

    @staticmethod
    def from_node(node):
        slotID = node.get('TIME_SLOT_ID')
        value = node.get('TIME_VALUE')
        if value is not None:
            return TimeSlot(slotID, int(node.get('TIME_VALUE')), xml_node=node)
        else:
            return TimeSlot(slotID, xml_node=node)

    @staticmethod
    def from_ts(ts, ID=None):
        value = ts2sec(ts) * 1000
        return TimeSlot(ID=ID, value=value)


class ELANAnnotation(DataObject):
    """ An ELAN abstract annotation (for both alignable and non-alignable annotations)
    """

    def __init__(self, ID, value, cve_ref=None, **kwargs):
        super().__init__(**kwargs)
        self.ID = ID
        self.value = value
        self.cve_ref = cve_ref

    @property
    def text(self):
        ''' An alias to ELANAnnotation.value '''
        return self.value

    def __repr__(self):
        return f"ELANAnnotation(ID={repr(self.ID)},value={repr(self.value)})"

    def __str__(self):
        return str(self.value)


class ELANTimeAnnotation(ELANAnnotation):
    """ An ELAN time-alignable annotation
    """
    def __init__(self, ID, from_ts, to_ts, value, xml_node=None, **kwargs):
        super().__init__(ID, value, **kwargs)
        self.from_ts = from_ts
        self.to_ts = to_ts
        self.__xml_node = xml_node

    @property
    def duration(self):
        return self.to_ts.sec - self.from_ts.sec

    def overlap(self, other):
        ''' Calculate overlap score between two time annotations
        Score = 0 means adjacent, score > 0 means overlapped, score < 0 means no overlap (the distance between the two)
        '''
        return min(self.to_ts, other.to_ts) - max(self.from_ts, other.from_ts)

    def __repr__(self):
        return '[{} -- {}] {}'.format(self.from_ts, self.to_ts, self.value)

    def __str__(self):
        return str(self.value)


class ELANRefAnnotation(ELANAnnotation):
    """ An ELAN ref annotation (not time alignable)
    """

    def __init__(self, ID, ref, previous, value, xml_node=None, **kwargs):
        super().__init__(ID, value, **kwargs)
        self.ref = ref  # ANNOTATION_REF
        self.previous = previous  # PREVIOUS_ANNOTATION
        self.__xml_node = xml_node


class LinguisticType(DataObject):
    def __init__(self, xml_node=None):
        """
        Linguistic Tier Type
        """
        data = {k.lower(): v for k, v in xml_node.attrib.items()} if xml_node is not None else {}
        if "time_alignable" in data:
            data["time_alignable"] = data["time_alignable"] == "true"
        super().__init__(**data)
        self.vocab = None
        self.tiers = []

    @property
    def ID(self):
        return self.linguistic_type_id

    def __repr__(self):
        return f"LinguisticType(ID={repr(self.ID)}, constraints={repr(self.constraints)}"


class ELANTier(DataObject):
    """ Represents an ELAN annotation tier """

    NONE = "None"
    TIME_SUB = "Time_Subdivision"
    SYM_SUB = "Symbolic_Subdivision"
    INCL = "Included_In"
    SYM_ASSOC = "Symbolic_Association"

    def __init__(self, type_ref_id, participant, ID, doc=None, default_locale=None, parent_ref=None, xml_node=None, **kwargs):
        """
        ELAN Tier Model which contains annotation objects
        """
        super().__init__(**kwargs)
        self.__type_ref = None
        self.__type_ref_id = type_ref_id
        self.participant = participant if participant else ''
        self.ID = ID
        self.default_locale = default_locale
        self.parent_ref = parent_ref
        self.parent = None
        self.doc = doc
        self.children = []
        self.annotations = []
        self.__xml_node = xml_node

    @property
    def time_alignable(self):
        """ Check if this tier contains time alignable annotations """
        return self.linguistic_type and self.linguistic_type.time_alignable

    @property
    def linguistic_type(self) -> LinguisticType:
        """ Linguistic type object of this Tier """
        return self.__type_ref

    def _set_type_ref(self, type_ref_object):
        """ [Internal function] Update type_ref object of this Tier """
        self.__type_ref = type_ref_object

    @property
    def _type_ref_id(self):
        return self.__type_ref_id

    def __getitem__(self, key):
        return self.annotations[key]

    def __iter__(self):
        return iter(self.annotations)

    def get_child(self, ID):
        ''' Get a child tier by ID, return None if nothing is found '''
        for child in self.children:
            if child.ID == ID:
                return child
        return None

    def filter(self, from_ts=None, to_ts=None):
        ''' Filter utterances by from_ts or to_ts or both
        If this tier is not a time-based tier everything will be returned
        '''
        for ann in self.annotations:
            if from_ts is not None and ann.from_ts is not None and ann.from_ts < from_ts:
                continue
            elif to_ts is not None and ann.to_ts is not None and ann.from_ts > to_ts:
                continue
            else:
                yield ann

    def __len__(self):
        return len(self.annotations)

    def __repr__(self):
        return 'Tier(ID={})'.format(self.ID)

    def __str__(self):
        return f'Tier(ID={repr(self.ID)}),type={repr(self.linguistic_type)})'.format(self.ID, self.linguistic_type)

    def add_alignable_annotation_xml(self, alignable):
        ann_id = alignable.get('ANNOTATION_ID')
        from_ts_id = alignable.get('TIME_SLOT_REF1')
        cve_ref = alignable.get('CVE_REF')  # controlled vocab ref
        if from_ts_id not in self.doc.time_order:
            raise ValueError("Time slot ID not found ({})".format(from_ts_id))
        else:
            from_ts = self.doc.time_order[from_ts_id]
        to_ts_id = alignable.get('TIME_SLOT_REF2')
        if to_ts_id not in self.doc.time_order:
            raise ValueError("Time slot ID not found ({})".format(to_ts_id))
        else:
            to_ts = self.doc.time_order[to_ts_id]
        # [TODO] ensure that from_ts < to_ts
        value_node = alignable.find('ANNOTATION_VALUE')
        if value_node is None:
            raise ValueError("ALIGNABLE_ANNOTATION node must contain an ANNOTATION_VALUE node")
        else:
            value = value_node.text if value_node.text else ''
            anno = ELANTimeAnnotation(ann_id, from_ts, to_ts, value, cve_ref=cve_ref, xml_node=alignable)
            self.annotations.append(anno)
            return anno

    def add_ref_annotation_xml(self, ref_node):
        ann_id = ref_node.get('ANNOTATION_ID')
        ref = ref_node.get('ANNOTATION_REF')
        previous = ref_node.get('PREVIOUS_ANNOTATION')
        cve_ref = ref_node.get('CVE_REF')  # controlled vocab ref
        value_node = ref_node.find('ANNOTATION_VALUE')
        if value_node is None:
            raise ValueError("REF_ANNOTATION node must contain an ANNOTATION_VALUE node")
        else:
            value = value_node.text if value_node.text else ''
            anno = ELANRefAnnotation(ann_id, ref, previous, value, cve_ref=cve_ref, xml_node=ref_node)
            self.annotations.append(anno)
            return anno

    def _add_annotation_xml(self, annotation_node) -> ELANAnnotation:
        """ [Internal function] Create an annotation from a node
        
        General users should not use this function.
        """
        alignable = annotation_node.find('ALIGNABLE_ANNOTATION')
        if alignable is not None:
            return self.add_alignable_annotation_xml(alignable)
        else:
            ref_ann_node = annotation_node.find('REF_ANNOTATION')
            if ref_ann_node is not None:
                return self.add_ref_annotation_xml(ref_ann_node)
            else:
                raise ValueError("ANNOTATION node must not be empty")


class ELANCVEntry(DataObject):

    ''' A controlled vocabulary entry '''
    
    def __init__(self, ID, lang_ref, value, description=None, **kwargs):
        super().__init__(**kwargs)
        self.ID = ID
        self.lang_ref = lang_ref
        self.value = value
        self.description = description

    def __repr__(self):
        return f'ELANCVEntry(ID={repr(self.ID)}, lang_ref={repr(self.lang_ref)}, value={repr(self.value)})'

    def __str__(self):
        return self.value


class ELANVocab(DataObject):
    ''' ELAN Controlled Vocabulary '''
    def __init__(self, ID, description, lang_ref, entries=None, **kwargs):
        super().__init__(**kwargs)
        self.ID = ID
        self.description = description
        self.lang_ref = lang_ref
        self.entries = list(entries) if entries else []
        self.entries_map = {e.ID: e for e in self.entries}
        self.tiers = []

    def __getitem__(self, key):
        return self.entries_map[key]

    def __iter__(self):
        return iter(self.entries)

    def __repr__(self):
        if self.description:
            return f'Vocab(ID={repr(self.ID)}, description={repr(self.description)})'
        else:
            return f'Vocab(ID={repr(self.ID)})'

    def __str__(self):
        return repr(self)

    @staticmethod
    def from_xml(node):
        CVID = node.get('CV_ID')
        description = ""
        lang_ref = ""
        entries = []
        for child in node:
            if child.tag == 'DESCRIPTION':
                description = child.text
                lang_ref = child.get('LANG_REF')
            elif child.tag == 'CV_ENTRY_ML':
                entryID = child.get('CVE_ID')
                entry_value_node = child.find('CVE_VALUE')
                entry_lang_ref = entry_value_node.get('LANG_REF')
                entry_value = entry_value_node.text
                entry_description = entry_value_node.get('DESCRIPTION')
                cv_entry = ELANCVEntry(entryID, entry_lang_ref, entry_value, description=entry_description)
                entries.append(cv_entry)
        return ELANVocab(CVID, description, lang_ref, entries=entries)


class ELANContraint(DataObject):
    """ ELAN Tier Constraints """

    def __init__(self, xml_node=None):
        super().__init__()
        if xml_node is not None:
            self.description = xml_node.get('DESCRIPTION')
            self.stereotype = xml_node.get('STEREOTYPE')


TierTuple = Tuple[ELANTier]
LinguisticTypeTuple= Tuple[LinguisticType]
ConstraintTuple = Tuple[ELANContraint]
VocabTuple = Tuple[ELANVocab]


class ELANDoc(DataObject):

    ''' This class represents an ELAN file (\*.eaf)
    '''
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.properties = OrderedDict()
        self.time_order = OrderedDict()
        self.__tiers_map = OrderedDict()  # internal - map tierIDs to tier objects
        self.__linguistic_types = []
        self.__constraints = []
        self.__vocabs = []
        self.__roots = []
        self.__xml_root = None
        self.__xml_header_node = None
        self.__xml_time_order_node = None

    @property
    def roots(self) -> TierTuple:
        """ All root-level tiers in this ELAN doc """
        return tuple(self.__roots)

    @property
    def vocabs(self) -> VocabTuple:
        """ A tuple of all existing controlled vocabulary objects in this ELAN file """
        return tuple(self.__vocabs)

    @property
    def constraints(self) -> ConstraintTuple:
        """ A tuple of all existing constraints in this ELAN file """
        return tuple(self.__constraints)

    @property
    def linguistic_types(self) -> LinguisticTypeTuple:
        """ A tuple of all existing linguistic types in this ELAN file """
        return tuple(self.__linguistic_types)

    def get_linguistic_type(self, type_id):
        ''' Get linguistic type by ID. Return None if can not be found '''
        for lingtype in self.__linguistic_types:
            if lingtype.linguistic_type_id == type_id:
                return lingtype
        return None

    def get_vocab(self, vocab_id):
        ''' Get controlled vocab list by ID '''
        for vocab in self.__vocabs:
            if vocab.ID == vocab_id:
                return vocab
        return None

    def get_participant_map(self):
        ''' Map participants to tiers
        Return a map from participant name to a list of corresponding tiers
        '''
        par_map = dd(list)
        for t in self.tiers():
            par_map[t.participant].append(t)
        return par_map

    def __getitem__(self, tierID):
        ''' Find a tier object using tierID '''
        return self.__tiers_map[tierID]
    
    def __iter__(self):
        """ Iterate through all tiers in this ELAN file """
        return iter(self.__tiers_map.values())

    def tiers(self) -> TierTuple:
        """ Collect all existing Tier in this ELAN file 
        """
        return tuple(self.__tiers_map.values())

    def _update_info_xml(self, node):
        ''' [Internal function] Update ELAN file metadata from an XML node 
        
        General users should not use this function.
        '''
        self.author = node.get('AUTHOR')
        self.date = node.get('DATE')
        self.fileformat = node.get('FORMAT')
        self.version = node.get('VERSION')

    def _update_header_xml(self, node):
        ''' [Internal function] Read ELAN doc information from a HEADER XML node 

        General users should not use this function.
        '''
        self.__xml_header_node = node
        self.media_file = node.get('MEDIA_FILE')
        self.time_units = node.get('TIME_UNITS')
        # extract media information
        media_node = node.find('MEDIA_DESCRIPTOR')
        if media_node is not None:
            self.media_url = media_node.get('MEDIA_URL')
            self.mime_type = media_node.get('MIME_TYPE')
            self.relative_media_url = media_node.get('RELATIVE_MEDIA_URL')
        # extract properties
        for prop_node in node.findall('PROPERTY'):
            self.properties[prop_node.get('NAME')] = prop_node.text

    def _add_tier_xml(self, tier_node) -> ELANTier:
        ''' [Internal function] Parse a TIER XML node, create an ELANTier object and link it to this ELANDoc

        General users should not use this function.
        '''
        type_ref = tier_node.get('LINGUISTIC_TYPE_REF')
        participant = tier_node.get('PARTICIPANT')
        tier_id = tier_node.get('TIER_ID')
        parent_ref = tier_node.get('PARENT_REF')
        default_locale = tier_node.get('DEFAULT_LOCALE')
        tier = ELANTier(type_ref, participant, tier_id, doc=self, default_locale=default_locale, parent_ref=parent_ref, xml_node=tier_node)
        # add child annotations
        for elem in tier_node:
            tier._add_annotation_xml(elem)
        if tier_id in self.__tiers_map:
            raise ValueError("Duplicated tier ID ({})".format(tier_id))
        self.__tiers_map[tier_id] = tier
        if tier.parent_ref is None:
            self.__roots.append(tier)
        return tier

    def _add_timeslot_xml(self, timeslot_node):
        ''' [Internal function] Parse a TimeSlot XML node and link it to current ELANDoc

        General users should not use this function.
        '''
        timeslot = TimeSlot.from_node(timeslot_node)
        self.time_order[timeslot.ID] = timeslot

    def _add_linguistic_type_xml(self, elem):
        ''' [Internal function] Parse a LinguisticType XML node and link it to current ELANDoc

        General users should not use this function.
        '''
        self.__linguistic_types.append(LinguisticType(elem))

    def _add_constraint_xml(self, elem):
        ''' [Internal function] Parse a CONSTRAINT XML node and link it to current ELANDoc

        General users should not use this function.
        '''
        self.__constraints.append(ELANContraint(elem))

    def _add_vocab_xml(self, elem):
        ''' [Internal function] Parse a CONTROLLED_VOCABULARY XML node and link it to current ELANDoc

        General users should not use this function.
        '''
        self.__vocabs.append(ELANVocab.from_xml(elem))

    def to_csv_rows(self) -> CSVTable:
        ''' Convert this ELANDoc into a CSV-friendly structure (i.e. list of list of strings) 
        
        :return: A list of list of strings
        :rtype: CSVTable
        '''
        rows = []
        for tier in self.tiers():
            for anno in tier.annotations:
                _from_ts = f"{anno.from_ts.sec:.3f}" if anno.from_ts else ''
                _to_ts = f"{anno.to_ts.sec:.3f}" if anno.to_ts else ''
                _duration = f"{anno.duration:.3f}" if anno.duration else ''
                rows.append((tier.ID, tier.participant, _from_ts, _to_ts, _duration, anno.value))
        return rows

    def to_xml_bin(self, encoding='utf-8', default_namespace=None, short_empty_elements=True, *args, **kwargs):
        ''' Generate EAF content in XML format

        :returns: EAF content
        :rtype: bytes
        '''
        _content = etree.tostring(self.__xml_root, encoding=encoding, method="xml",
                                  short_empty_elements=short_empty_elements, *args, **kwargs)
        return _content

    def save(self, path, encoding='utf-8', xml_declaration=None,
             default_namespace=None, short_empty_elements=True, *args, **kwargs):
        ''' Write ELANDoc to an EAF file '''
        _content = self.to_xml_bin(encoding=encoding,
                                   xml_declaration=xml_declaration,
                                   default_namespace=default_namespace,
                                   short_empty_elements=short_empty_elements)
        chio.write_file(path, _content, encoding=encoding)

    @classmethod
    def open_eaf(cls, eaf_path, encoding='utf-8', *args, **kwargs):
        with chio.open(eaf_path, encoding=encoding, *args, **kwargs) as eaf_stream:
            return cls.parse_eaf_stream(eaf_stream)

    @classmethod
    def parse_eaf_stream(cls, eaf_stream):
        _root = etree.fromstring(eaf_stream.read())
        _doc = ELANDoc()
        _doc.__xml_root = _root
        _doc._update_info_xml(_root)
        for elem in _root:
            if elem.tag == 'HEADER':
                _doc._update_header_xml(elem)
            elif elem.tag == 'TIME_ORDER':
                _doc.__xml_time_order_node = elem
                for time_elem in elem:
                    _doc._add_timeslot_xml(time_elem)
            elif elem.tag == 'TIER':
                _doc._add_tier_xml(elem)
            elif elem.tag == 'LINGUISTIC_TYPE':
                _doc._add_linguistic_type_xml(elem)
            elif elem.tag == 'CONSTRAINT':
                _doc._add_constraint_xml(elem)
            elif elem.tag == 'CONTROLLED_VOCABULARY':
                _doc._add_vocab_xml(elem)
            else:
                logging.getLogger(__name__).warning(f"Ignored element type -- {elem.tag}")
        # linking parts together
        # linguistic_types -> vocabs
        for lingtype in _doc.linguistic_types:
            if lingtype.controlled_vocabulary_ref:
                lingtype.vocab = _doc.get_vocab(lingtype.controlled_vocabulary_ref)
        # resolves tiers' roots, parents, and type
        for tier in _doc.tiers():
            lingtype = _doc.get_linguistic_type(tier._type_ref_id)
            tier._set_type_ref(lingtype)
            lingtype.tiers.append(tier)  # type -> tiers
            if lingtype.vocab:
                lingtype.vocab.tiers.append(tier)  # vocab -> tiers
            if tier.parent_ref is not None:
                tier.parent = _doc[tier.parent_ref]
                _doc[tier.parent_ref].children.append(tier)
        return _doc


open_eaf = ELANDoc.open_eaf
parse_eaf_stream = ELANDoc.parse_eaf_stream


def __resolve(elan_doc):
    ''' Ensure that everything is linked together (e.g. tiers, vocabs, etc.)
    '''
    # link linguistic_types -> vocabs
    for lingtype in elan_doc.linguistic_types:
        if lingtype.controlled_vocabulary_ref:
            lingtype.vocab = elan_doc.get_vocab(lingtype.controlled_vocabulary_ref)
    # resolves tiers' roots, parents, and type
    for tier in elan_doc.tiers():
        lingtype = elan_doc.get_linguistic_type(tier._type_ref_id)
        tier._set_type_ref(lingtype)
        lingtype.tiers.append(tier)  # type -> tiers
        if lingtype.vocab:
            lingtype.vocab.tiers.append(tier)  # vocab -> tiers
        if tier.parent_ref is not None:
            tier.parent = elan_doc[tier.parent_ref]
            elan_doc[tier.parent_ref].children.append(tier)


def _legacy_parse_eaf_stream(eaf_stream):
    ''' Parse an EAF text stream and return an ELAN object

    :param eaf_stream: a text-based stream object
    :type eaf_stream: Text I/O
    :return: an ELANDoc object
    :rtype: speach.elan.ELANDoc
    '''
    elan_doc = ELANDoc()
    current_tier = None
    for event, elem in etree.iterparse(eaf_stream, events=('start', 'end')):
        if event == 'start':
            if elem.tag == 'ANNOTATION_DOCUMENT':
                elan_doc._update_info_xml(elem)
            elif elem.tag == 'TIER':
                current_tier = elan_doc._add_tier_xml(elem)
        else:  # event == 'end'
            if elem.tag == 'HEADER':
                elan_doc._update_header_xml(elem)
                elem.clear()  # no need to keep header node in memory
            elif elem.tag == 'TIME_SLOT':
                elan_doc._add_timeslot_xml(elem)
                elem.clear()
            elif elem.tag == 'ANNOTATION':
                current_tier._add_annotation_xml(elem)
                elem.clear()
            elif elem.tag == 'LINGUISTIC_TYPE':
                elan_doc._add_linguistic_type_xml(elem)
                elem.clear()
            elif elem.tag == 'CONSTRAINT':
                elan_doc._add_constraint_xml(elem)
                elem.clear()
            elif elem.tag == 'CONTROLLED_VOCABULARY':
                elan_doc._add_vocab_xml(elem)
                elem.clear()
    __resolve(elan_doc)  # link parts together
    return elan_doc


def _legacy_open_eaf(eaf_path, encoding='utf-8', *args, **kwargs) -> ELANDoc:
    ''' Read and parse an EAF file and return an ELANDoc object 

    :param eaf_path: Path to an EAF file
    :return: An ELANDoc object
    :rtype: speach.elan.ELANDoc
    '''
    
    with chio.open(eaf_path, encoding=encoding, *args, **kwargs) as eaf_stream:
        elan_doc = _legacy_parse_eaf_stream(eaf_path)
        return elan_doc



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Test basic TTL SQLite
'''

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import os
import unittest
import logging

from speach import ttl
from speach.sqlite import TTLSQLite


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

TEST_DIR = os.path.dirname(os.path.realpath(__file__))


def getLogger():
    return logging.getLogger(__name__)


def get_db(use_ram=True, auto_clear=True):
    if use_ram:
        return TTLSQLite(':memory:')
    else:
        db_loc = os.path.join(TEST_DIR, 'data', 'test.db')
        if auto_clear:
            if os.path.isfile(db_loc):
                os.unlink(db_loc)
        return TTLSQLite(db_loc)


# -------------------------------------------------------------------------------
# Tests
# -------------------------------------------------------------------------------

class TestTTLSQLite(unittest.TestCase):

    def test_extra_fields(self):
        doc = ttl.Document(corpusID=2)
        self.assertEqual(doc.corpusID, 2)

    def test_read_json(self):
        docjson = os.path.join(TEST_DIR, 'data', 'test.json')
        doc = ttl.read_json(docjson)
        getLogger().debug(doc[0].to_json())
        self.assertTrue(len(doc))

    def test_ident(self):
        db = get_db(True)
        with db.ctx() as ctx:
            corpus = db.new_corpus('eng', ctx=ctx)
            doc = db.new_doc(name='eng1', title='eng1', lang='en', corpusID=corpus.ID, ctx=ctx)
            author = db.get_doc_meta_by_key(doc.name, 'author', ctx=ctx)
            # there must be no meta as of now
            self.assertIsNone(author)
            db.set_doc_meta(doc.name, 'author', 'Le Tuan Anh', ctx=ctx)
            # select it out
            doc1 = db.doc.by_id(1, ctx=ctx)
            author = db.get_doc_meta_by_key(doc1.name, 'author', ctx=ctx)
            self.assertEqual(author.value, 'Le Tuan Anh')
            # make a sentence with ident
            sent = doc1.new_sent('It rains.', ident='10000')
            sent.ID = None
            db.save_sent(sent, ctx=ctx)
            # select sentence by ident
            sid = ctx.sent.select('ident = ?', ('10000',))[0].ID
            sent = db.get_sent(sid, ctx=ctx)
            self.assertEqual(sent.text, 'It rains.')

    def test_store_objects(self):
        db = get_db(True)
        with db.ctx() as ctx:
            getLogger().info("The database should be empty right now")
            self.assertFalse(ctx.tag.select())
            # create a sample corpus
            corpus = db.new_corpus('jpn', ctx=ctx)
            title = 'Japanese Corpus 1'
            corpus.title = title
            ctx.corpus.save(corpus)
            self.assertEqual(ctx.corpus.by_id(1).title, corpus.title)
            corpus = db.corpus.select_single('ID=?', (1,), ctx=ctx)
            self.assertEqual(corpus.title, title)
            # create a document
            doc = db.new_doc(name='jpn1', title='日本語', lang='jp', corpusID=corpus.ID, ctx=ctx)
            doc1 = db.doc.by_id(1, ctx=ctx)
            self.assertEqual(doc1.name, 'jpn1')
            self.assertEqual(doc1.ID, 1)
            # add sentences from JSON
            testdoc_path = os.path.join(TEST_DIR, 'data', 'test.json')
            docjson = ttl.read_json(testdoc_path)
            docjson[0].comment = 'mikeneko ga suki desu.'
            docjson[0].new_tag('I like calico cats.', tagtype='eng')
            docjson[0][0].new_tag('mi', tagtype='romaji')
            docjson[1].comment = 'It rains.'
            docjson[1].new_tag('It rains.', tagtype='eng')
            docjson[1][0].new_tag('ame', tagtype='romaji')
            docjson[2].comment = 'The girl eats the cake.'
            docjson[2].new_tag('The girl eats the cake.', tagtype='eng')
            docjson[2][0].new_tag('onna no ko', tagtype='romaji')
            for sent in docjson:
                sent.ID = None
                sent.docID = doc.ID
                db.save_sent(sent, ctx=ctx)
            # test retrieve data back
            self.assertEqual(len(ctx.sent.select()), 3)
            sent = db.get_sent(1, ctx=ctx)
            self.assertTrue(sent)
            self.assertTrue(sent.tags)
            self.assertTrue(sent.tokens)
            self.assertTrue(sent.concepts)
            self.assertEqual(sent.get_tag('eng').label, 'I like calico cats.')
            self.assertEqual(sent.comment, 'mikeneko ga suki desu.')
            getLogger().debug(sent.tags)
            getLogger().debug(sent.tokens)
            getLogger().debug(sent.concepts)
            # test lexicon
            lex = list(db.lexicon(ctx=ctx))
            self.assertEqual(len(lex), 14)
            # use limit
            lex = [(t, c) for t, c in db.lexicon(limit=2, ctx=ctx)]
            self.assertEqual(lex, [('。', 3), ('が', 2)])


class TestTTLSQLiteMeta(unittest.TestCase):

    def test_collection_meta(self):
        db = get_db(True)
        with db.ctx() as ctx:
            metas = db.get_meta(ctx=ctx)
            self.assertFalse(len(metas))
            # insert metas
            db.set_meta(key='name', value='foo', ctx=ctx)
            db.set_meta(key='search', value='bing', ctx=ctx)
            metas = db.get_meta(ctx=ctx)
            actual = set((x.key, x.value) for x in metas)
            expected = {('search', 'bing'), ('name', 'foo')}
            self.assertEqual(expected, actual)
            name = db.get_meta_by_key('name', ctx=ctx)
            self.assertEqual(name.value, 'foo')
            # set meta with existing key -> update
            db.set_meta('search', 'yahoo', ctx=ctx)
            search = db.get_meta_by_key('search', ctx=ctx)
            self.assertEqual(search.value, 'yahoo')

    def test_doc_meta(self):
        db = get_db(True)
        with db.ctx() as ctx:
            metas = db.get_doc_meta('eng1', ctx=ctx)
            self.assertFalse(len(metas))
            # insert metas
            db.set_doc_meta(name='eng1', key='name', value='foo', ctx=ctx)
            db.set_doc_meta(name='eng1', key='search', value='bing', ctx=ctx)
            db.set_doc_meta(name='vie1', key='search', value='yahoo', ctx=ctx)
            metas = db.get_doc_meta(name='eng1', ctx=ctx)
            actual = set((x.name, x.key, x.value) for x in metas)
            expected = {('eng1', 'search', 'bing'), ('eng1', 'name', 'foo')}
            metas = db.get_doc_meta(name='vie1', ctx=ctx)
            actual = set((x.name, x.key, x.value) for x in metas)
            expected = {('vie1', 'search', 'yahoo')}
            self.assertEqual(expected, actual)
            name = db.get_doc_meta_by_key('eng1', 'name', ctx=ctx)
            self.assertEqual(name.value, 'foo')
            # set meta with existing key -> update
            db.set_doc_meta('eng1', 'search', 'yahoo', ctx=ctx)
            search = db.get_doc_meta_by_key('eng1', 'search', ctx=ctx)
            self.assertEqual(search.value, 'yahoo')
            db.set_doc_meta('vie1', 'search', 'wolfram-alpha', ctx=ctx)
            search = db.get_doc_meta_by_key('vie1', 'search', ctx=ctx)
            self.assertEqual(search.value, 'wolfram-alpha')

    def test_cor_meta(self):
        db = get_db(True)
        with db.ctx() as ctx:
            metas = db.get_cor_meta('eng', ctx=ctx)
            self.assertFalse(len(metas))
            # insert metas
            db.set_cor_meta(name='eng', key='name', value='foo', ctx=ctx)
            db.set_cor_meta(name='eng', key='search', value='bing', ctx=ctx)
            db.set_cor_meta(name='vie', key='search', value='yahoo', ctx=ctx)
            metas = db.get_cor_meta(name='eng', ctx=ctx)
            actual = set((x.name, x.key, x.value) for x in metas)
            expected = {('eng', 'search', 'bing'), ('eng', 'name', 'foo')}
            metas = db.get_cor_meta(name='vie', ctx=ctx)
            actual = set((x.name, x.key, x.value) for x in metas)
            expected = {('vie', 'search', 'yahoo')}
            self.assertEqual(expected, actual)
            name = db.get_cor_meta_by_key('eng', 'name', ctx=ctx)
            self.assertEqual(name.value, 'foo')
            # set meta with existing key -> update
            db.set_cor_meta('eng', 'search', 'yahoo', ctx=ctx)
            search = db.get_cor_meta_by_key('eng', 'search', ctx=ctx)
            self.assertEqual(search.value, 'yahoo')
            db.set_cor_meta('vie', 'search', 'wolfram-alpha', ctx=ctx)
            search = db.get_cor_meta_by_key('vie', 'search', ctx=ctx)
            self.assertEqual(search.value, 'wolfram-alpha')
            # select all meta
            metas = ctx.meta_cor.select()
            expected = [{'name': 'eng', 'key': 'name', 'value': 'foo'}, {'name': 'eng', 'key': 'search', 'value': 'yahoo'}, {'name': 'vie', 'key': 'search', 'value': 'wolfram-alpha'}]
            actual = [m.to_dict() for m in metas]
            self.assertEqual(expected, actual)


# -------------------------------------------------------------------------------
# MAIN
# -------------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

'''
speach - a Python library for managing, annotating, and converting natural language corpuses
using popular formats (CoNLL, ELAN, Praat, CSV, JSON, SQLite, VTT, Audacity, TTL, TIG, ISF)
'''

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.


from .__version__ import __author__, __email__, __copyright__, __maintainer__
from .__version__ import __credits__, __license__, __description__, __url__
from .__version__ import __version_major__, __version_long__, __version__, __status__

from chirptext import ttl
from .sqlite import TTLSQLite


__all__ = ['ttl', 'TTLSQLite',
           "__version__", "__author__", "__description__", "__copyright__"]



# -*- coding: utf-8 -*-

'''
TTL Interlinear Gloss (TIG) format support

More information: https://en.wikipedia.org/wiki/Interlinear_gloss
An interlinear text will commonly consist of some or all of the following, usually in this order, from top to bottom:
    The original orthography (typically in italic or bold italic),
    a conventional transliteration into the Latin alphabet,
    a phonetic transcription,
    a morphophonemic transliteration,
    a word-by-word or morpheme-by-morpheme gloss, where morphemes within a word are separated by hyphens or other punctuation,
    a free translation, which may be placed in a separate paragraph or on the facing page if the structures of the languages are too different for it to follow the text line by line.
'''
# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.


import re
import logging
from difflib import ndiff
from collections import OrderedDict
import warnings

from chirptext import DataObject, piter
from chirptext import chio
from chirptext.deko import is_kana, parse
from chirptext import ttl



# ----------------------------------------------------------------------
# Configuration
# ----------------------------------------------------------------------

def getLogger():
    return logging.getLogger(__name__)


# ----------------------------------------------------------------------
# Models
# ----------------------------------------------------------------------

# Source: https://en.wikipedia.org/wiki/Interlinear_gloss
# An interlinear text will commonly consist of some or all of the following, usually in this order, from top to bottom:
#     The original orthography (typically in italic or bold italic),
#     a conventional transliteration into the Latin alphabet,
#     a phonetic transcription,
#     a morphophonemic transliteration,
#     a word-by-word or morpheme-by-morpheme gloss, where morphemes within a word are separated by hyphens or other punctuation,
#     a free translation, which may be placed in a separate paragraph or on the facing page if the structures of the languages are too different for it to follow the text line by line.
class IGRow(DataObject):
    def __init__(self, text='', transliteration='', transcription='', morphtrans='', morphgloss='', wordgloss='', translation='', **kwargs):
        """
        """
        super().__init__()
        self.text = text
        self.transliteration = transliteration
        self.transcription = transcription
        self.morphtrans = morphtrans
        self.morphgloss = morphgloss
        self.wordgloss = wordgloss
        self.translation = translation
        self.update(kwargs)

    def to_ttl(self):
        ttl_sent = ttl.Sentence(text=self.text)
        data = self.to_dict()
        for l in TTLIG.KNOWN_LABELS:
            if l not in ['text', 'orth', 'tokens'] and l in data and data[l]:
                ttl_sent.new_tag(data[l], tagtype=l)
        if self.tokens:
            _tokens = parse_ruby(self.tokens)
            ttl_sent.tokens = (t.text() for t in _tokens)
            for ttl_token, furi_token in zip(ttl_sent, _tokens):
                if furi_token.surface != furi_token.text():
                    ttl_token.new_tag(furi_token.surface, tagtype='furi')
            if self.morphtrans:
                _morphtokens = tokenize(self.morphtrans)
                if len(_morphtokens) != len(ttl_sent):
                    getLogger().warning("Morphophonemic transliteration line and tokens line are mismatched for sentence: {}".format(self.ident or self.ID or self.Id or self.id or self.text))
                else:
                    for t, m in zip(ttl_sent, _morphtokens):
                        t.new_tag(m, tagtype='mtrans')
            if self.pos:
                _postokens = tokenize(self.pos)
                if len(_postokens) != len(ttl_sent):
                    getLogger().warning("Part-of-speech line and tokens line are mismatched for sentence: {}".format(self.ident or self.ID or self.Id or self.id or self.text))
                else:
                    for t, m in zip(ttl_sent, _postokens):
                        t.pos = m
            if self.lemma:
                _lemmas = tokenize(self.lemma)
                if len(_lemmas) != len(ttl_sent):
                    getLogger().warning("Lemma line and tokens line are mismatched for sentence: {}".format(self.ident or self.ID or self.Id or self.id or self.text))
                else:
                    for t, m in zip(ttl_sent, _lemmas):
                        t.lemma = m
            if self.morphgloss:
                _glosstokens = tokenize(self.morphgloss)
                if len(_glosstokens) != len(ttl_sent):
                    getLogger().warning("morpheme-by-morpheme gloss and tokens lines are mismatched for sentence {}".format(self.ident or self.ID or self.Id or self.id or self.text))
                else:
                    for t, m in zip(ttl_sent, _glosstokens):
                        t.new_tag(m, tagtype='mgloss')
            if self.wordgloss:
                _glosstokens = tokenize(self.wordgloss)
                if len(_glosstokens) != len(ttl_sent):
                    getLogger().warning("word-by-word gloss and tokens lines are mismatched for sentence {}".format(self.ident or self.ID or self.Id or self.id or self.text))
                else:
                    for t, m in zip(ttl_sent, _glosstokens):
                        t.new_tag(m, tagtype='wgloss')
        return ttl_sent

    def to_expex(self, default_ident=''):
        lines = []
        sent_ident = self.ident if self.ident else default_ident
        lines.append('\\ex \\label{{{}}}'.format(sent_ident))
        lines.append('\\begingl[aboveglftskip=0pt]')
        tags = ['gla', 'glb', 'glc']
        # process tokens and gloss
        glosses = []
        lengths = []
        if self.tokens:
            lengths.append(make_expex_gloss(self.tokens, glosses, tags.pop(0)))
        if self.morphtrans:
            lengths.append(make_expex_gloss(self.morphtrans, glosses, tags.pop(0)))
        if self.morphgloss:
            lengths.append(make_expex_gloss(self.morphgloss, glosses, tags.pop(0)))
        if self.concept:
            if tags:
                lengths.append(make_expex_gloss(self.concept, glosses, tags.pop(0)))
            else:
                getLogger().warning("There are too many gloss lines in sentence {}. {}".format(sent_ident, self.text))
        # ensure that number of tokens are the same
        if len(lengths) > 1:
            for line_len in lengths[1:]:
                if line_len != lengths[0]:
                    getLogger().warning("Inconsistent tokens and morphgloss for sentence {}. {} ({} v.s {})".format(sent_ident, self.text, line_len, lengths[0]))
                    break
        lines.extend(glosses)
        lines.append('\\glft \lit{{{}}}//'.format(escape_latex(self.text)))
        lines.append('\\endgl')
        lines.append('\\xe')
        return '\n'.join(lines)

    # Matrix alias
    @property
    def orth(self):
        return self.text

    @orth.setter
    def orth(self, value):
        self.text = value

    @property
    def translit(self):
        return self.transliteration

    @translit.setter
    def translit(self, value):
        self.transliteration = value

    @property
    def translat(self):
        return self.translation

    @translat.setter
    def translat(self, value):
        self.translation = value

    @property
    def gloss(self):
        return self.morphgloss

    @gloss.setter
    def gloss(self, value):
        self.morphgloss = value

    @property
    def tsduration(self):
        if self.tsfrom is None or self.tsto is None:
            return None
        else:
            return self.tsto - self.tsfrom
        
    def overlap(self, other):
        ''' Calculate overlap score between this utterance and another
        Score = 0 means adjacent, score > 0 means overlapped, score < 0 means no overlap (the distance between the two)
        '''
        return min(self.tsto, other.tsto) - max(self.tsfrom, other.tsfrom)



LATEX_SPECIAL_CHARS = P = re.compile('([%${_#&}])')


def escape_latex(text):
    text = LATEX_SPECIAL_CHARS.sub(r"\\\g<1>", text.replace('\\', '\\textbackslash '))
    text = text.replace('<', '\\textless ').replace('>', '\\textgreater ')
    return text if text.strip() else '{}'


def make_expex_gloss(raw, lines, gloss_tag):
    _tokens = tokenize(raw)
    lines.append('\\{} {} //'.format(gloss_tag, ' '.join((escape_latex(t) for t in _tokens))))
    return len(_tokens)


class TTLIG(object):

    # default lines
    AUTO_LINES = ['tokens', 'morphtrans', 'gloss']
    MANUAL_TAG = '__manual__'
    AUTO_TAG = '__auto__'
    ROBUST_TAG = '__robust__'
    SPECIAL_LABELS = [AUTO_TAG, MANUAL_TAG, ROBUST_TAG]
    KNOWN_META = ['language', 'language code', 'lines', 'author', 'date']
    ANNOTATIONS = ['flag', 'font', 'font-global']
    SPECIAL_FEATURES = ['furigana', 'furi']
    CORPUS_MANAGEMENT = ['comment', 'source', 'vetted', 'judgement', 'phenomena', 'url', 'type']
    SYNTAX = ['tree', 'dtree']
    SEMANTICS = ['concept', 'mrs', 'dmrs', 'predicates', 'preds']
    DISCOURSE = ['tsfrom', 'tsto', 'speaker']
    INTERLINEAR_GLOSS = ['ident', 'orth', 'morphgloss', 'wordgloss', 'translation', 'text', 'translit', 'translat', 'tokens', 'lemma', 'pos']
    KNOWN_LABELS = AUTO_LINES + KNOWN_META + ANNOTATIONS + SPECIAL_FEATURES + CORPUS_MANAGEMENT + SYNTAX + SEMANTICS + INTERLINEAR_GLOSS + DISCOURSE
    # [TODO] Add examples & description for each of these labels

    def __init__(self, meta):
        self.meta = meta

    def row_format(self):
        if 'Lines' in self.meta:
            lines = self.meta['Lines'].strip()
            if lines:
                return self.meta['Lines'].strip().split()
        return []

    def _parse_row(self, line_list, line_tags):
        if not line_list:
            raise ValueError("Lines cannot be empty")
        if not line_tags or line_tags == [TTLIG.AUTO_TAG]:
            # auto
            # first line = text, last line = translation, transli
            if len(line_list) == 1:
                return IGRow(text=line_list[0])
            elif len(line_list) == 2:
                return IGRow(text=line_list[0], translation=line_list[-1])
            else:
                others = {k: v for k, v in zip(TTLIG.AUTO_LINES, line_list[1:-1])}
                return IGRow(text=line_list[0], translation=line_list[-1], **others)
        elif line_tags == [TTLIG.MANUAL_TAG]:
            line_dict = {}
            for line in line_list:
                tag_idx = line.find(':')
                if tag_idx == -1:
                    raise ValueError("Invalid line (no tag found) -> {}".format(line))
                _tag = line[:tag_idx].strip()
                _val = line[tag_idx + 1:].lstrip().rstrip('\r\n')
                if _tag.lower() not in TTLIG.KNOWN_LABELS:
                    getLogger().warning("Unknown tag was used ({}): {}".format(_tag, _val))
                line_dict[_tag] = _val
            return IGRow(**line_dict)
        else:
            # explicit, just zip them
            if len(line_tags) != len(line_list):
                raise ValueError("Mismatch number of lines for {} - {}".format(line_tags, line_list))
            return IGRow(**{k: v for k, v in zip(line_tags, line_list)})

    def read_iter(self, stream):
        line_tags = self.row_format()
        for tag in line_tags:
            if tag.lower() not in TTLIG.KNOWN_LABELS + TTLIG.SPECIAL_LABELS:
                getLogger().warning("Unknown label in header: {}".format(tag))
        for row in IGStreamReader._iter_stream(stream):
            yield self._parse_row(row, line_tags)


class IGStreamReader(object):

    META_LINE = re.compile(r'(?P<key>[\w\s]+):\s*(?P<value>.*)')

    @staticmethod
    def _read_header(ig_stream):
        ''' Read the TTLIG header from a stream '''
        lines = piter(ig_stream)
        first = next(lines)
        if first.strip() != '# TTLIG':
            raise Exception("Invalid TTLIG header. TTLIG files must start with # TTLIG")
        meta = OrderedDict()
        for line in lines:
            if line.startswith("#"):
                continue
            m = IGStreamReader.META_LINE.match(line)
            if m:
                key = m.group('key').strip()
                value = m.group('value')
                if key in meta:
                    getLogger().warning("Key {} is duplicated in the header".format(key))
                meta[key] = value
            else:
                # this line is weird
                break
            if not lines.peep() or not lines.peep().value.strip():
                # if next line is empty, break
                break
        return meta

    @staticmethod
    def _iter_stream(ig_stream):
        lines = piter(ig_stream)
        current = []
        for line_raw in lines:
            line = line_raw.rstrip('\r\n')
            if not line.startswith('#') and line:
                # not a comment or an empty line
                current.append(line)
            if not line or not lines.peep() or not lines.peep().value:
                if current:
                    yield current
                    current = []


def read_stream_iter(ttlig_stream):
    meta = IGStreamReader._read_header(ttlig_stream)
    ig_obj = TTLIG(meta)
    return ig_obj.read_iter(ttlig_stream)


def read_stream(ttlig_stream):
    ''' read TTLIG stream '''
    return [s for s in read_stream_iter(ttlig_stream)]


def read(ttlig_filepath):
    ''' Read TTLIG file '''
    with chio.open(ttlig_filepath, mode='rt') as infile:
        return read_stream(infile)


FURIMAP = re.compile(r'\{(?P<text>[\w%％]+?)/(?P<furi>[\w%％]+?)\}')


class RubyToken(DataObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'groups' not in kwargs:
            self.groups = []

    def append(self, group):
        self.groups.append(group)

    def text(self):
        return ''.join(str(x) for x in self.groups)

    def to_html(self):
        frags = []
        for g in self.groups:
            if isinstance(g, RubyFrag):
                frags.append(g.to_html())
            else:
                frags.append(str(g))
        return ''.join(frags)

    def to_code(self):
        frags = []
        for g in self.groups:
            if isinstance(g, RubyFrag):
                frags.append('{{{text}/{furi}}}'.format(text=g.text, furi=g.furi))
            else:
                frags.append(str(g))
        return ''.join(frags)

    def to_anki(self):
        ''' Export token to Anki fugigana format '''
        frags = []
        for g in self.groups:
            if isinstance(g, RubyFrag):
                frags.append('{text}[{furi}]'.format(text=g.text, furi=g.furi))
            else:
                frags.append(str(g))
        return ''.join(frags)
    
    def __str__(self):
        return self.text()

    @staticmethod
    def from_furi(surface, kana):
        ruby = RubyToken(surface=surface)
        if is_kana(surface):
            ruby.append(surface)
            return ruby
        edit_seq = ndiff(surface, kana)
        kanji = ''
        text = ''
        furi = ''
        before = ''
        expected = ''
        for item in edit_seq:
            if item.startswith('- '):
                # flush text if needed
                if expected and kanji and furi:
                    ruby.append(RubyFrag(text=kanji, furi=furi))
                    kanji = ''
                    furi = ''
                if text:
                    ruby.append(text)
                    text = ''
                kanji += item[2:]
            elif item.startswith('+ '):
                if expected and item[2:] == expected:
                    if expected and kanji and furi:
                        ruby.append(RubyFrag(text=kanji, furi=furi))
                        kanji = ''
                        furi = ''
                    ruby.append(item[2:])
                    expected = ''
                else:
                    furi += item[2:]
            elif item.startswith('  '):
                if before == '-' and not furi:
                    # shifting happened
                    expected = item[2:]
                    furi += item[2:]
                else:
                    text += item[2:]
                    # flush if possible
                    if kanji and furi:
                        ruby.append(RubyFrag(text=kanji, furi=furi))
                        kanji = ''
                        furi = ''
                    else:
                        # possible error?
                        pass
            before = item[0]  # end for
        # flush final parts
        if kanji:
            if furi:
                ruby.append(RubyFrag(text=kanji, furi=furi))
            else:
                ruby.append(kanji)
        elif text:
            ruby.append(text)
        return ruby


class RubyFrag(DataObject):
    def __init__(self, text, furi, **kwargs):
        super().__init__(text=text, furi=furi, **kwargs)

    def __repr__(self):
        return "Ruby(text={}, furi={})".format(repr(self.text), repr(self.furi))

    def to_html(self):
        return "<ruby><rb>{}</rb><rt>{}</rt></ruby>".format(self.text, self.furi)

    def __str__(self):
        return self.text if self.text else ''


def parse_furigana(text):
    ''' Parse TTLRuby token (returns a RubyToken)'''
    if text is None:
        raise ValueError
    start = 0
    ruby = RubyToken(surface=text)
    ms = [(m.groupdict(), m.span()) for m in FURIMAP.finditer(text)]
    # frag: ruby fragment
    for frag, (cfrom, cto) in ms:
        if start < cfrom:
            ruby.append(text[start:cfrom])
        ruby.append(RubyFrag(text=frag['text'], furi=frag['furi']))
        start = cto
    if start < len(text):
        ruby.append(text[start:len(text)])
    return ruby


class TTLTokensParser(object):
    ''' TTL Tokens parser '''
    def __init__(self, escapechar='\\', delimiter=' '):
        self.escapechar = escapechar
        self.delimiter = delimiter

    def parse(self, text):
        tokens = []
        current = []
        chars = piter(text)
        is_escaping = False
        for c in chars:
            if is_escaping:
                current.append(c)
                is_escaping = False
            elif c == self.escapechar:
                # add the next character to current token
                if not chars.peep():
                    raise ValueError("Escape char ({}) cannot be the last character".format(self.escapechar))
                elif chars.peep() and chars.peep().value not in (self.escapechar, self.delimiter):
                    getLogger().warning("Escape char ({}) should not be used for normal character ({}). This can be a potential bug in the data.".format(self.escapechar, chars.peep().value))
                is_escaping = True
            elif c == self.delimiter:
                # flush
                if current:
                    tokens.append(''.join(current))
                    current = []
                # else -> ignore
            else:
                current.append(c)
            # flush if current is the last character
            if chars.peep() is None and current:
                tokens.append(''.join(current))
                current = []
        return tokens

    def parse_ruby(self, text):
        ''' Return a list of RubyToken '''
        return [parse_furigana(t) for t in self.parse(text)]


DEFAULT_TTL_TOKEN_PARSER = TTLTokensParser()


def tokenize(text):
    return DEFAULT_TTL_TOKEN_PARSER.parse(text)


def parse_ruby(text):
    return DEFAULT_TTL_TOKEN_PARSER.parse_ruby(text)


def make_ruby_html(text):
    tokens = parse_ruby(text)
    return DEFAULT_TTL_TOKEN_PARSER.delimiter.join(r.to_html() for r in tokens)


def mctoken_to_furi(token):
    ''' Convert mecab token to TTLIG format '''
    warnings.warn("mctoken_to_furi() is deprecated and will be removed in near future. Use RubyToken.from_furi() instead", DeprecationWarning, stacklevel=2)
    return RubyToken.from_furi(token.surface, token.reading_hira())


def text_to_igrow(txt):
    ''' Parse text to TTLIG format '''
    msent = parse(txt)
    tokens = []
    pos = []
    lemmas = []
    for token in msent:
        if token.is_eos:
            continue
        pos.append(token.pos3())
        r = RubyToken.from_furi(token.surface, token.reading_hira())
        tokens.append(r.to_code())
        lemmas.append(token.root)
    igrow = IGRow(text=txt, tokens=' '.join(tokens), pos=' '.join(pos), lemma=' '.join(lemmas))
    return igrow



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Setup script for speach

Latest version can be found at https://github.com/neocl/speach

@author: Le Tuan Anh <tuananh.ke@gmail.com>
@license: MIT
'''

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import io
from setuptools import setup


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


readme_file = 'README.md'
long_description = read(readme_file)
pkg_info = {}
exec(read('speach/__version__.py'), pkg_info)

with open('requirements.txt', 'r') as infile:
    requirements = infile.read().splitlines()

setup(
    name='speach',
    version=pkg_info['__version__'],
    url=pkg_info['__url__'],
    project_urls={
        "Bug Tracker": "https://github.com/neocl/speach/issues",
        "Source Code": "https://github.com/neocl/speach/"
    },
    keywords="nlp",
    license=pkg_info['__license__'],
    author=pkg_info['__author__'],
    tests_require=requirements + ['coverage'],
    install_requires=requirements,
    author_email=pkg_info['__email__'],
    description=pkg_info['__description__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['speach',
              'speach.data'],
    package_data={'speach': ['data/*.sql', 'data/*.gz']},
    include_package_data=True,
    platforms='any',
    test_suite='test',
    # Reference: https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=['Programming Language :: Python',
                 'Development Status :: 2 - Pre-Alpha',
                 'License :: OSI Approved :: {}'.format(pkg_info['__license__']),
                 'Environment :: Plugins',
                 'Intended Audience :: Education',
                 'Intended Audience :: Science/Research',
                 'Intended Audience :: Information Technology',
                 'Intended Audience :: Developers',
                 'Operating System :: OS Independent',
                 'Topic :: Text Processing',
                 'Topic :: Text Processing :: Linguistic',
                 'Topic :: Software Development :: Libraries :: Python Modules']
)



# -*- coding: utf-8 -*-

"""
TTL Tools
"""

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import os
import logging
from lxml import etree

from chirptext import TextReport, FileHelper
from chirptext import chio
from chirptext.cli import CLIApp, setup_logging

from speach import ttl, TTLSQLite, ttlig, orgmode
from speach.elan import parse_eaf_stream

# ----------------------------------------------------------------------
# Configuration
# ----------------------------------------------------------------------

try:
    setup_logging('logging.json', 'logs')
except Exception:
    pass


def getLogger():
    return logging.getLogger(__name__)


FORMAT_TTL = 'ttl'
FORMAT_EXPEX = 'expex'


# ----------------------------------------------------------------------
# Functions
# ----------------------------------------------------------------------

def get_doc_length(name, ctx):
    query = 'SELECT COUNT(*) FROM sentence WHERE docID = (SELECT ID FROM document WHERE name=?)'
    return ctx.select_scalar(query, (name,))


def make_db(cli, args):
    ''' Convert TTL-TXT to TTL-SQLite '''
    print("Reading document ...")
    ttl_doc = ttl.Document.read_ttl(args.ttl)
    print("Sentences: {}".format(len(ttl_doc)))
    db = TTLSQLite(args.db)
    db_corpus = db.ensure_corpus(name=args.corpus)
    db_doc = db.ensure_doc(name=args.doc if args.doc else ttl_doc.name, corpus=db_corpus)
    if get_doc_length(db_doc.name, ctx=db.ctx()):
        print("Document is not empty, program aborted.")
    else:
        # insert sents
        with db.ctx() as ctx:
            ctx.buckmode()
            ctx.execute('BEGIN;')
            for idx, sent in enumerate(ttl_doc):
                if args.topk and args.topk <= idx:
                    break
                print("Processing sent #{}/{}".format(idx + 1, len(ttl_doc)))
                sent.ID = None
                sent.docID = db_doc.ID
                db.save_sent(sent, ctx=ctx)
            ctx.execute('COMMIT;')
    print("Done!")


def process_tig(cli, args):
    ''' Convert TTLIG file to TTL format '''
    if args.format == FORMAT_TTL:
        sc = 0
        ttl_writer = ttl.TxtWriter.from_path(args.output) if args.output else None
        with chio.open(args.ttlig) as infile:
            for sent in ttlig.read_stream_iter(infile):
                sc += 1
                if ttl_writer is not None:
                    ttl_sent = sent.to_ttl()
                    ttl_writer.write_sent(ttl_sent)
        if ttl_writer is not None:
            print("Output file: {}".format(args.output))
        print("Processed {} sentence(s).".format(sc))
    elif args.format == FORMAT_EXPEX:
        sc = 0
        output = TextReport(args.output)
        output.print("\\newcommand{\\lit}[1]{``#1''}     %%% literal meaning")
        output.print()
        output.print()
        output.print()
        with chio.open(args.ttlig) as infile:
            for idx, sent in enumerate(ttlig.read_stream_iter(infile)):
                sc += 1
                output.print(sent.to_expex(default_ident=idx + 1))
                output.print()
                output.print()
                output.print()
    else:
        print("Format {} is not supported".format(args.format))


def jp_line_proc(line, iglines):
    igrow = ttlig.text_to_igrow(line.replace('\u3000', ' ').strip())
    iglines.append(igrow.text)
    iglines.append(igrow.tokens)
    iglines.append("")


def convert_org_to_tig(inpath, outpath):
    title, meta, lines = orgmode.read(inpath)
    meta.append(("Lines", "text tokens"))
    out = orgmode.org_to_ttlig(title, meta, lines, jp_line_proc)
    output = TextReport(outpath)
    for line in out:
        output.print(line)


def org_to_ttlig(cli, args):
    ''' Convert ORG file to TTLIG format '''
    if args.orgfile:
        # single file mode
        convert_org_to_tig(args.orgfile, args.output)
    elif args.orgdir:
        if not args.output:
            print("Output directory is required for batch mode")
            exit()
        # make output directory
        if not os.path.exists(args.output):
            print("Make directory: {}".format(args.output))
            os.makedirs(args.output)
        else:
            print("Output directory: {}".format(args.output))
        filenames = FileHelper.get_child_files(args.orgdir)
        for filename in filenames:
            infile = os.path.join(args.orgdir, filename)
            outfile = os.path.join(args.output, FileHelper.replace_ext(filename, 'tig'))
            if os.path.exists(outfile):
                print("File {} exists. SKIPPED".format(outfile))
            else:
                print("Generating: {} => {}".format(infile, outfile))
                convert_org_to_tig(infile, outfile)
    print("Done")


def make_text(sent, delimiter=' '):
    frags = []
    if sent.tokens:
        for tk in sent:
            furi = tk.find('furi', default=None)
            if furi:
                frags.append(ttlig.make_ruby_html(furi.label))
            else:
                frags.append(tk.text)
    html_text = delimiter.join(frags) if frags else sent.text
    return "<text>{}</text>".format(html_text)


def make_html(cli, args):
    ''' Convert TTL to HTML '''
    print("Reading document ...")
    ttl_doc = ttl.Document.read_ttl(args.ttl)
    output = TextReport(args.output)
    doc_node = etree.Element('doc')
    for sent in ttl_doc:
        sent_node = etree.SubElement(doc_node, 'sent')
        text_node = etree.XML(make_text(sent, delimiter=args.delimiter))
        sent_node.append(text_node)
        if sent.get_tag('translation'):
            etree.SubElement(sent_node, 'br')
            trans_node = etree.SubElement(sent_node, 'trans')
            trans_node.text = sent.get_tag('translation').label
        etree.SubElement(sent_node, 'br')
        etree.SubElement(sent_node, 'br')
    output.write(etree.tostring(doc_node, encoding='unicode', pretty_print=not args.compact))


def sec_str(a_float):
    return "{:.3f}".format(a_float)


def convert_eaf_to_csv(eaf_path, csv_path):
    with chio.open(eaf_path) as eaf_stream:
        elan = parse_eaf_stream(eaf_stream)
        rows = elan.to_csv_rows()
        chio.write_tsv(csv_path, rows, quoting=chio.QUOTE_MINIMAL)


def eaf_to_csv(cli, args):
    ''' Convert ELAN file (*.eaf) to TSV (Tab-separated Values) format '''
    convert_eaf_to_csv(args.eaf, args.output)
    print("Output has been written to: {}".format(args.output))


# -------------------------------------------------------------------------------
# Main
# -------------------------------------------------------------------------------

def main():
    ''' Speach toolkit - managing, annotating, and converting natural language corpuses '''
    app = CLIApp(desc='Speach tools', logger=__name__)
    # add tasks
    task = app.add_task('convert', func=make_db)
    task.add_argument('ttl', help='TTL file')
    task.add_argument('db', help='TTL DB file')
    task.add_argument('corpus', help='Corpus name')
    task.add_argument('doc', help='Document name', default=None)
    task.add_argument('-k', '--topk', help='Only select the top k frequent elements', default=None, type=int)

    task = app.add_task('ig', func=process_tig)
    task.add_argument('ttlig', help='TTLIG file')
    task.add_argument('-o', '--output', help='Output TTL file')
    task.add_argument('-f', '--format', help='Output format', choices=[FORMAT_EXPEX, FORMAT_TTL], default=FORMAT_TTL)

    task = app.add_task('org', func=org_to_ttlig)
    task.add_argument('-f', '--orgfile', help='ORG file')
    task.add_argument('-d', '--orgdir', help='ORG directory (batch mode)')
    task.add_argument('-o', '--output', help='Output TTL file or directory')

    task = app.add_task('html', func=make_html)
    task.add_argument('ttl', help='TTL file')
    task.add_argument('-o', '--output', help='path to output HTML file')
    task.add_argument('-c', '--compact', help='Do not use pretty print', action='store_true')
    task.add_argument('-d', '--delimiter', help='Token delimiter', default=' ')

    task = app.add_task('eaf2csv', func=eaf_to_csv)
    task.add_argument('eaf', help='Input EAF file')
    task.add_argument('-o', '--output', help='path to output CSV file')

    # run app
    app.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Test WebVTT support
"""

# This code is a part of speach library: https://github.com/neocl/speach/
# :copyright: (c) 2018 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import os
import unittest
import logging

from speach import vtt


# -------------------------------------------------------------------------------
# Configuration
# -------------------------------------------------------------------------------

TEST_DIR = os.path.dirname(os.path.realpath(__file__))


def getLogger():
    return logging.getLogger(__name__)


# -------------------------------------------------------------------------------
# Tests
# -------------------------------------------------------------------------------

class TestTTLIG(unittest.TestCase):

    def test_wrong_input(self):
        self.assertRaises(ValueError, lambda: vtt.sec2ts(None))
        self.assertRaises(ValueError, lambda: vtt.sec2ts(-5))
        self.assertRaises(ValueError, lambda: vtt.sec2ts("-5"))
        self.assertRaises(ValueError, lambda: vtt.sec2ts("5.a"))
        # ts2sec
        self.assertRaises(ValueError, lambda: vtt.ts2sec(None))
        self.assertRaises(ValueError, lambda: vtt.ts2sec('00:00:00.00'))  # wrong msec
        self.assertRaises(TypeError, lambda: vtt.ts2sec(self))

    def test_vtt_ts_util(self):
        print("Test WebVTT timestamp conversion functions")
        inp = 0
        expected_ts = '00:00:00.000'
        ts = vtt.sec2ts(inp)
        out = vtt.ts2sec(ts)
        self.assertEqual(ts, expected_ts)
        self.assertEqual(out, inp)

        inp = 26
        expected_ts = '00:00:26.000'
        ts = vtt.sec2ts(inp)
        out = vtt.ts2sec(ts)
        self.assertEqual(ts, expected_ts)
        self.assertEqual(out, inp)

        inp = "26.999"
        expected_ts = '00:00:26.999'
        ts = vtt.sec2ts(inp)
        out = vtt.ts2sec(ts)
        self.assertEqual(ts, expected_ts)
        self.assertEqual(out, float(inp))

        inp = 61
        expected_ts = '00:01:01.000'
        ts = vtt.sec2ts(inp)
        out = vtt.ts2sec(ts)
        self.assertEqual(ts, expected_ts)
        self.assertEqual(out, inp)

        inp = 83.027000
        expected_ts = '00:01:23.027'
        ts = vtt.sec2ts(inp)
        out = vtt.ts2sec(ts)
        self.assertEqual(ts, expected_ts)
        self.assertEqual(out, inp)

        inp = 186.963
        expected_ts = '00:03:06.963'
        ts = vtt.sec2ts(inp)
        out = vtt.ts2sec(ts)
        self.assertEqual(ts, expected_ts)
        self.assertEqual(out, inp)

        inp = 7497.999
        expected_ts = '02:04:57.999'
        ts = vtt.sec2ts(inp)
        out = vtt.ts2sec(ts)
        self.assertEqual(ts, expected_ts)
        self.assertEqual(out, inp)

        # more than 99 hours
        inp = 997497.999
        expected_ts = '277:04:57.999'
        ts = vtt.sec2ts(inp)
        out = vtt.ts2sec(ts)
        self.assertEqual(ts, expected_ts)
        self.assertEqual(out, inp)


# -------------------------------------------------------------------------------
# MAIN
# -------------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()
